#include <cryptoauthlib.h>
#include "basic/atca_basic.h"

#define BOOL_TO_ONOFF(b) (b ? "on" : "off")
#define NIBBLE_TO_HEX_CHAR(i) ((i <= 9) ? ('0' + i) : ('A' - 10 + i))
#define HIGH_NIBBLE(i) ((i >> 4) & 0x0F)
#define LOW_NIBBLE(i) (i & 0x0F)

#define HEX_CHAR_TO_NIBBLE(c) ((c >= 'A') ? (c - 'A' + 0x0A) : (c - '0'))
#define HEX_PAIR_TO_BYTE(h, l) ((HEX_CHAR_TO_NIBBLE(h) << 4) + HEX_CHAR_TO_NIBBLE(l))

#define IS_HEX_CHAR(c) ( (((c >= 'A') && (c <= 'F')) || ((c >= '0') && (c <= '9'))) ? true : false)
#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))

#define debugSerial SerialUSB
#define debugBaud 57600
#define debugDelay 10000

// Configs from /tests/atca_test.c
uint8_t test_ecc608_configdata[ATCA_ECC_CONFIG_SIZE] = {
    0x01, 0x23, 0x00, 0x00, 0x00, 0x00, 0x60, 0x00, 0x04, 0x05, 0x06, 0x07, 0xEE, 0x01, 0x01, 0x00,  //15
    0xC0, 0x00, 0xA1, 0x00, 0x8F, 0x2F, 0xC4, 0x44, 0x87, 0x20, 0xC4, 0xF4, 0x8F, 0x0F, 0x0F, 0x0F,  //31, 5
    0x9F, 0x8F, 0x83, 0x64, 0xC4, 0x44, 0xC4, 0x64, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F,  //47
    0x0F, 0x0F, 0x0F, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF,  //63
    0x00, 0x00, 0x00, 0x00, 0xFF, 0x84, 0x03, 0xBC, 0x09, 0x69, 0x76, 0x00, 0x00, 0x00, 0x00, 0x00,  //79
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x0E, 0x40, 0x00, 0x00, 0x00, 0x00,  //95
    0x33, 0x00, 0x1C, 0x00, 0x13, 0x00, 0x1C, 0x00, 0x3C, 0x00, 0x3E, 0x00, 0x1C, 0x00, 0x33, 0x00,  //111
    0x1C, 0x00, 0x1C, 0x00, 0x38, 0x10, 0x30, 0x00, 0x3C, 0x00, 0x3C, 0x00, 0x32, 0x00, 0x30, 0x00   //127
};

uint8_t test_ecc_configdata[ATCA_ECC_CONFIG_SIZE] = {
    0x01, 0x23, 0x00, 0x00, 0x00, 0x00, 0x50, 0x00, 0x04, 0x05, 0x06, 0x07, 0xEE, 0x00, 0x01, 0x00, //15
    0xC0, 0x00, 0x55, 0x00, 0x8F, 0x2F, 0xC4, 0x44, 0x87, 0x20, 0xC4, 0xF4, 0x8F, 0x0F, 0x8F, 0x8F, //31, 5
    0x9F, 0x8F, 0x83, 0x64, 0xC4, 0x44, 0xC4, 0x64, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, 0x0F, //47
    0x0F, 0x0F, 0x0F, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, //63
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, //79
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //95
    0x33, 0x00, 0x1C, 0x00, 0x13, 0x00, 0x1C, 0x00, 0x3C, 0x00, 0x1C, 0x00, 0x1C, 0x00, 0x33, 0x00, //111
    0x1C, 0x00, 0x1C, 0x00, 0x3C, 0x00, 0x30, 0x00, 0x3C, 0x00, 0x3C, 0x00, 0x32, 0x00, 0x30, 0x00  //127
};

void setup() 
{
  while ((!debugSerial) && (millis() < debugDelay)){
    // Wait debugDelay milliseconds for debugSerial to open
  }

  // Start debug stream
  debugSerial.begin(debugBaud);
  debugSerial.println("Starting...");
  debugSerial.println();
  
  //Init chip
  cfg_ateccx08a_i2c_default.iface.atcai2c.bus = 0;
  debugSerial.print("Initialising ATECCX08A:...");
  showResult(atcab_init(&cfg_ateccx08a_i2c_default));
  debugSerial.println();

  // Read configuration lock status
  bool isLocked = false;
  debugSerial.print("Reading configuration lock status:...");
  showResult(atcab_is_locked(LOCK_ZONE_CONFIG, &isLocked));
  debugSerial.print("The configuration zone is:...");
  debugSerial.println((isLocked ? "LOCKED" : "UNLOCKED"));
  debugSerial.println();

  // We can only configure and lock the device once ever
  if (!isLocked) {
    // Read revision information
    uint8_t rev_info[4];
    debugSerial.print("Reading revision info:...");
    showResult(atcab_info(rev_info));
    printHex(rev_info, sizeof(rev_info), sizeof(rev_info));
    debugSerial.println();

    // Select config based on revision info
    uint8_t* CONFIG = NULL;
    if (rev_info[2] == 0x50) {
      CONFIG = test_ecc_configdata;
    }
    else if (rev_info[2] == 0x60) {
      CONFIG = test_ecc608_configdata;
    }

    if (CONFIG == NULL) {
      // Halt if chip is revision is not recognised
      debugSerial.println("Error: Unknown chip variant...");
      halt();
    }
    else {
      // Write configuration
      debugSerial.print("Writing configuration zone:...");
      showResult(atcab_write_config_zone(CONFIG));
      debugSerial.println();
  
      // Read back configuration
      uint8_t read_data[ATCA_ECC_CONFIG_SIZE];
      debugSerial.print("Reading configuration zone:...");
      showResult(atcab_read_config_zone(read_data));
      debugSerial.println();
      
      // Lock config zone
      debugSerial.print("Locking configuration zone:...");
      showResult(atcab_lock_config_zone());
      debugSerial.println();
  
      // Read data zone lock status
      isLocked = false;
      debugSerial.print("Reading data zone lock status:...");
      showResult(atcab_is_locked(LOCK_ZONE_DATA, &isLocked));
      debugSerial.print("The data zone is:...");
      debugSerial.println((isLocked ? "LOCKED" : "UNLOCKED"));
      debugSerial.println();
    
      //Lock data zone if unlocked
      if (!isLocked) {
        debugSerial.print("Locking data zone:...");
        showResult(atcab_lock_data_zone());
        debugSerial.println();
      }
    }
  }
  else {
    // Device is already configured and locked
    debugSerial.println("Device is already configured...");
    debugSerial.println();
  }
}

void loop() 
{ 
}

void printHex(const uint8_t* buff, uint8_t len, uint8_t width)
{
    char strA[16];
    for (uint8_t i = 0; i < len; i++)
    {
        if (i % width == 0)
        {
          sprintf(strA,"0x%04x",i);
          debugSerial.print(String((i == 0 ? "" : "\r\n")) + String(strA) + ": ");
        }
        sprintf(strA,"0x%02X, ",buff[i]);
        debugSerial.print(strA);
        
    }
    debugSerial.println();
}

void printRawHex(const uint8_t* buff, uint8_t len)
{
    for (uint8_t i = 0; i < len; i++)
    {
      debugSerial.print((char)NIBBLE_TO_HEX_CHAR(HIGH_NIBBLE(buff[i])));
      debugSerial.print((char)NIBBLE_TO_HEX_CHAR(LOW_NIBBLE(buff[i])));
    }
    debugSerial.println();
}

void halt()
{
  while (true) {
    delay(1000);
  }
}

void showResult(ATCA_STATUS result)
{
  String message;
  
  switch(result) {
    case ATCA_SUCCESS                 :message = "SUCCESS"; break;
    case ATCA_CONFIG_ZONE_LOCKED      :message = "CONFIG_ZONE_LOCKED"; break;
    case ATCA_DATA_ZONE_LOCKED        :message = "DATA_ZONE_LOCKED"; break;
    case ATCA_WAKE_FAILED             :message = "WAKE_FAILED"; break;
    case ATCA_CHECKMAC_VERIFY_FAILED  :message = "CHECKMAC_VERIFY_FAILED"; break;
    case ATCA_PARSE_ERROR             :message = "PARSE_ERROR"; break;
    case ATCA_STATUS_CRC              :message = "STATUS_CRC"; break;
    case ATCA_STATUS_UNKNOWN          :message = "STATUS_UNKNOWN"; break;
    case ATCA_STATUS_ECC              :message = "STATUS_ECC"; break;
    case ATCA_FUNC_FAIL               :message = "FUNC_FAIL"; break;
    case ATCA_GEN_FAIL                :message = "GEN_FAIL"; break;
    case ATCA_BAD_PARAM               :message = "BAD_PARAM"; break;
    case ATCA_INVALID_ID              :message = "INVALID_ID"; break;
    case ATCA_INVALID_SIZE            :message = "INVALID_SIZE"; break;
    case ATCA_RX_FAIL                 :message = "RX_FAIL"; break;
    case ATCA_RX_NO_RESPONSE          :message = "RX_NO_RESPONSE"; break;
    case ATCA_RESYNC_WITH_WAKEUP      :message = "RESYNC_WITH_WAKEUP"; break;
    case ATCA_PARITY_ERROR            :message = "PARITY_ERROR"; break;
    case ATCA_TX_TIMEOUT              :message = "TX_TIMEOUT"; break;
    case ATCA_RX_TIMEOUT              :message = "RX_TIMEOUT"; break;
    case ATCA_COMM_FAIL               :message = "COMM_FAIL"; break;
    case ATCA_TIMEOUT                 :message = "TIMEOUT"; break;
    case ATCA_BAD_OPCODE              :message = "BAD_OPCODE"; break;
    case ATCA_WAKE_SUCCESS            :message = "WAKE_SUCCESS"; break;
    case ATCA_EXECUTION_ERROR         :message = "EXECUTION_ERROR"; break;
    case ATCA_UNIMPLEMENTED           :message = "UNIMPLEMENTED"; break;
    case ATCA_ASSERT_FAILURE          :message = "ASSERT_FAILURE"; break;
    case ATCA_TX_FAIL                 :message = "TX_FAIL"; break;
    case ATCA_NOT_LOCKED              :message = "NOT_LOCKED"; break;
    case ATCA_NO_DEVICES              :message = "NO_DEVICES"; break;
  }

  debugSerial.println(message);

  if (result != ATCA_SUCCESS) {
    halt();
  }
}
